import time
import sys
import pathlib
import importlib
import traceback
import mlx.core as mx
import mlx.nn as nn
from typing import Union


########################################################
# Baseline
########################################################
class Model(nn.Module):
    """
    Model that performs a 2D convolution.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        kernel_size (Union[int, tuple]): Size of the convolution kernel.
        stride (Union[int, tuple]): Stride of the convolution. Default is 1.
    """

    def __init__(self, in_channels: int, out_channels: int, kernel_size: Union[int, tuple], stride: Union[int, tuple] = 1):
        super(Model, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, stride)

    def __call__(self, x):
        """
        Args:
            x (mx.array): Input tensor of shape (batch_size, height, width, in_channels).
        Returns:
            mx.array: Output tensor of shape (batch_size, height, width, out_channels).
        """
        return self.conv(x)


########################################################
# Weco Solution
########################################################
def load_module_from_path(module_path: str, add_to_sys_modules: bool = False):
    # Clean out all old compiled extensions to prevent namespace collisions during build
    module_path = pathlib.Path(module_path)
    name = module_path.stem
    spec = importlib.util.spec_from_file_location(name, module_path)
    mod = importlib.util.module_from_spec(spec)  # type: ignore
    if add_to_sys_modules:
        sys.modules[name] = mod
    spec.loader.exec_module(mod)  # type: ignore
    return mod


########################################################
# Benchmark
########################################################
def get_inputs(batch_size, img_height, img_width, img_channels):
    # MLX doesn't use device parameter like PyTorch, as it automatically uses Metal
    return mx.random.normal(shape=(batch_size, img_height, img_width, img_channels), dtype=mx.float32)


def bench(f, inputs, n_warmup, n_rep):
    # Warm up
    for _ in range(n_warmup):
        result = f(inputs)
        mx.eval(result)  # Force computation due to lazy evaluation

    t_avg = 0.0
    for _ in range(n_rep):
        # Clear cache before timing
        mx.clear_cache()

        start_time = time.time()
        result = f(inputs)
        mx.eval(result)  # Force computation
        mx.synchronize()  # Wait for all computations to complete
        t_avg += time.time() - start_time

    t_avg /= n_rep * 1e-3
    return t_avg


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--solution-path", type=str, required=True)
    args = parser.parse_args()

    # init and input parameters
    batch_size = 4
    img_height = 224
    img_width = 224
    img_channels = 3
    out_channels = 64
    kernel_size = 3
    stride = 1

    # Set the default device to 0
    mx.set_default_device(mx.gpu)

    # load solution module
    try:
        mx.random.seed(0)
        solution_module = load_module_from_path(args.solution_path, add_to_sys_modules=False)
        solution_model = solution_module.Model(img_channels, out_channels, kernel_size, stride)
        assert hasattr(solution_model, "__call__")
    except Exception:
        print(f"Candidate module initialization failed: {traceback.format_exc()}")
        exit(1)

    mx.random.seed(0)
    baseline_model = Model(img_channels, out_channels, kernel_size, stride)

    # measure correctness
    n_correctness_trials = 10
    max_diff_avg = 0
    for _ in range(n_correctness_trials):
        inputs = get_inputs(batch_size, img_height, img_width, img_channels)
        baseline_output = baseline_model(inputs)
        optimized_output = solution_model(inputs)
        max_diff = mx.max(mx.abs(optimized_output - baseline_output))
        mx.eval(max_diff)  # Force computation
        max_diff_avg += max_diff.item()  # Convert to Python scalar
    max_diff_avg /= n_correctness_trials
    print(f"max float diff between values of baseline and optimized model: {max_diff_avg}")

    # measure performance
    inputs = get_inputs(batch_size, img_height, img_width, img_channels)
    n_warmup = 1000
    n_rep = 5000

    # baseline
    t_avg_baseline = bench(baseline_model, inputs, n_warmup, n_rep)
    print(f"baseline time: {t_avg_baseline:.2f}ms")

    # optimized
    t_avg_optimized = bench(solution_model, inputs, n_warmup, n_rep)
    print(f"optimized time: {t_avg_optimized:.2f}ms")

    print(f"speedup: {t_avg_baseline / t_avg_optimized:.2f}x")



from pathlib import Path
import pandas as pd


def predict(test: Path, save: Path):
    # TODO: Add a model here

    test_data = pd.read_csv(test)
    submission = pd.DataFrame({"PassengerId": test_data["PassengerId"], "Transported": False})
    submission.to_csv(save, index=False)
    print(f"Test submission saved to {save}")


if __name__ == "__main__":
    # This block is primarily for testing the script directly,
    # it's not used by the weco evaluation loop.
    script_dir = Path(__file__).resolve().parent
    # Use validation data as test data *for this test block only*
    train_file_path = script_dir / "public" / "train.csv"
    print("train_file_path:", train_file_path)
    test_file_path = script_dir / "public" / "test.csv"
    print("test_file_path:", test_file_path)
    submission_output_path = script_dir / "submission.csv"

    # Call predict with the DataFrame and the correct output path
    predict(train_file_path, test_file_path, submission_output_path)
    print(f"Test submission saved to {submission_output_path}")



import time
import sys
import os
import pathlib
import importlib
import traceback
import torch
import torch.nn as nn


########################################################
# Baseline
########################################################
class Model(nn.Module):
    """
    Model that performs a matrix multiplication, division, summation, and scaling.
    """

    def __init__(self, input_size, hidden_size, scaling_factor):
        super(Model, self).__init__()
        self.weight = nn.Parameter(torch.randn(hidden_size, input_size))
        self.scaling_factor = scaling_factor

    def forward(self, x):
        """
        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, input_size).
        Returns:
            torch.Tensor: Output tensor of shape (batch_size, hidden_size).
        """
        x = torch.matmul(x, self.weight.T)
        x = x / 2
        x = torch.sum(x, dim=1, keepdim=True)
        x = x * self.scaling_factor
        return x


########################################################
# Weco Solution
########################################################
def load_module_from_path(module_path: str, add_to_sys_modules: bool = False):
    # Clean out all old compiled extensions to prevent namespace collisions during build
    module_path = pathlib.Path(module_path)
    name = module_path.stem
    spec = importlib.util.spec_from_file_location(name, module_path)
    mod = importlib.util.module_from_spec(spec)  # type: ignore
    if add_to_sys_modules:
        sys.modules[name] = mod
    spec.loader.exec_module(mod)  # type: ignore
    return mod


########################################################
# Benchmark
########################################################
os.environ["MAX_JOBS"] = "1"  # number of workers for building with ninja


def get_inputs(B, N, device):
    return torch.randn(B, N, device=device, dtype=torch.float32)


@torch.no_grad()
def bench(f, inputs, n_warmup, n_rep):
    # Warm up
    for _ in range(n_warmup):
        f(inputs)  # noqa

    # Benchmark
    device_type = inputs.device.type
    t_avg = 0.0
    for _ in range(n_rep):
        # Clear cache before timing
        if device_type == "cuda":
            torch.cuda.empty_cache()
        elif device_type == "mps":
            torch.mps.empty_cache()

        # time forward pass
        start_time = time.time()
        f(inputs)
        t_avg += time.time() - start_time

        # Synchronize after each iteration
        if device_type == "cuda":
            torch.cuda.synchronize()
        elif device_type == "mps":
            torch.mps.synchronize()

    t_avg /= n_rep * 1e-3
    return t_avg


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--solution-path", type=str, required=True)
    parser.add_argument("--device", default="cpu", type=str)
    args = parser.parse_args()

    # benchmark parameters
    n_correctness_trials = 10
    n_warmup = 1000
    n_rep = 5000

    # init and input parameters
    batch_size, input_size, hidden_size, scaling_factor = 128, 10, 20, 1.5

    # load solution module
    try:
        torch.manual_seed(0)
        solution_module = load_module_from_path(args.solution_path, add_to_sys_modules=False)
        solution_model = solution_module.Model(input_size, hidden_size, scaling_factor).to(args.device)
        assert isinstance(solution_model, nn.Module)
        assert hasattr(solution_model, "forward")
    except Exception:
        print(f"Candidate module initialization failed: {traceback.format_exc()}")
        exit(1)

    torch.manual_seed(0)
    baseline_model = Model(input_size, hidden_size, scaling_factor).to(args.device)

    # measure correctness
    max_diff_avg = 0
    for _ in range(n_correctness_trials):
        inputs = get_inputs(batch_size, input_size, args.device)
        baseline_output = baseline_model(inputs)
        optimized_output = solution_model(inputs)
        max_diff_avg += torch.max(torch.abs(optimized_output - baseline_output))
    max_diff_avg /= n_correctness_trials
    print(f"max float diff between values of baseline and optimized model: {max_diff_avg}")

    # measure performance
    inputs = get_inputs(batch_size, input_size, args.device)
    t_avg_baseline = bench(baseline_model, inputs, n_warmup, n_rep)
    print(f"baseline time: {t_avg_baseline:.2f}ms")
    t_avg_optimized = bench(solution_model, inputs, n_warmup, n_rep)
    print(f"optimized time: {t_avg_optimized:.2f}ms")
    print(f"speedup: {t_avg_baseline / t_avg_optimized:.2f}x")



from typing import Dict, Any
import rich
import requests
from weco import __pkg_version__, __base_url__
import sys


def handle_api_error(e: requests.exceptions.HTTPError, console: rich.console.Console) -> None:
    """Extract and display error messages from API responses in a structured format."""
    error_message = str(e)  # Default message
    console.print(f"[bold red]Error:[/] {error_message}")
    sys.exit(1)


def start_optimization_session(
    console: rich.console.Console,
    source_code: str,
    evaluation_command: str,
    metric_name: str,
    maximize: bool,
    steps: int,
    code_generator_config: Dict[str, Any],
    evaluator_config: Dict[str, Any],
    search_policy_config: Dict[str, Any],
    additional_instructions: str = None,
    api_keys: Dict[str, Any] = {},
    timeout: int = 800,
) -> Dict[str, Any]:
    """Start the optimization session."""
    with console.status("[bold green]Starting Optimization..."):
        try:
            response = requests.post(
                f"{__base_url__}/sessions",
                json={
                    "source_code": source_code,
                    "additional_instructions": additional_instructions,
                    "objective": {"evaluation_command": evaluation_command, "metric_name": metric_name, "maximize": maximize},
                    "optimizer": {
                        "steps": steps,
                        "code_generator": code_generator_config,
                        "evaluator": evaluator_config,
                        "search_policy": search_policy_config,
                    },
                    "metadata": {"client_name": "cli", "client_version": __pkg_version__, **api_keys},
                },
                timeout=timeout,
            )
            response.raise_for_status()
            return response.json()
        except requests.exceptions.HTTPError as e:
            handle_api_error(e=e, console=console)


def evaluate_feedback_then_suggest_next_solution(
    console: rich.console.Console,
    session_id: str,
    execution_output: str,
    additional_instructions: str = None,
    api_keys: Dict[str, Any] = {},
    timeout: int = 800,
) -> Dict[str, Any]:
    """Evaluate the feedback and suggest the next solution."""
    try:
        response = requests.post(
            f"{__base_url__}/sessions/{session_id}/suggest",
            json={
                "execution_output": execution_output,
                "additional_instructions": additional_instructions,
                "metadata": {**api_keys},
            },
            timeout=timeout,
        )
        response.raise_for_status()
        return response.json()
    except requests.exceptions.HTTPError as e:
        handle_api_error(e=e, console=console)


def get_optimization_session_status(
    console: rich.console.Console, session_id: str, include_history: bool = False, timeout: int = 800
) -> Dict[str, Any]:
    """Get the current status of the optimization session."""
    try:
        response = requests.get(
            f"{__base_url__}/sessions/{session_id}", params={"include_history": include_history}, timeout=timeout
        )
        response.raise_for_status()
        return response.json()
    except requests.exceptions.HTTPError as e:
        handle_api_error(e=e, console=console)
